import axios from "axios";
import type { AxiosInstance, AxiosBasicCredentials } from "axios";
import type { WorkflowModeOut } from "@/client/workflow";

export function requiredRepositoryFiles(modes: WorkflowModeOut[]) {
  const list = [
    "main.nf",
    "CHANGELOG.md",
    "README.md",
    "docs/usage.md",
    "docs/output.md",
  ];
  if (modes.length > 0) {
    list.push(...modes.map((mode) => mode.schema_path));
  } else {
    list.push("nextflow_schema.json");
  }
  return list;
}

export function determineGitIcon(repo_url?: string): string {
  let gitProvider = "git-alt";
  if (repo_url != null) {
    if (repo_url.includes("github")) {
      gitProvider = "github";
    } else if (repo_url.includes("gitlab")) {
      gitProvider = "gitlab";
    } else if (repo_url.includes("bitbucket")) {
      gitProvider = "bitbucket";
    }
  }
  return "fa-brands fa-".concat(gitProvider);
}

export abstract class GitRepository {
  protected repo: URL;
  protected gitCommitHash: string;
  protected token?: string;
  protected httpClient: AxiosInstance;

  protected constructor(
    repoUrl: string,
    gitCommitHash: string,
    token?: string,
  ) {
    this.repo = new URL(repoUrl);
    this.gitCommitHash = gitCommitHash;
    if (token != undefined && token.length > 0) {
      this.token = token;
    }
    this.httpClient = axios.create({
      maxRedirects: 2,
    });
  }

  async checkFileExist(filepath: string): Promise<boolean> {
    try {
      await this.httpClient.head(this.fileUrl(filepath));
    } catch (e) {
      return false;
    }
    return true;
  }

  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  async downloadFile(filepath: string): Promise<any> {
    try {
      return await this.httpClient.get(await this.downloadFileUrl(filepath));
    } catch (e) {
      return "";
    }
  }

  async checkFilesExist(
    files: string[],
    raiseError = false,
  ): Promise<boolean[]> {
    const checks = Promise.all(
      files.map((filepath) => this.checkFileExist(filepath)),
    );
    const results = await checks;
    if (raiseError) {
      const missingFiles = files.filter((file, index) => !results[index]);
      if (missingFiles.length > 0) {
        throw new Error(missingFiles.reduce((a, b) => a + "," + b));
      }
    }
    return results;
  }

  protected abstract fileUrl(filepath: string): string;

  protected abstract downloadFileUrl(filepath: string): Promise<string>;

  static buildRepository(
    repoUrl: string,
    gitCommitHash: string,
    token?: string,
  ): GitRepository {
    if (repoUrl.includes("github")) {
      return new GithubRepository(repoUrl, gitCommitHash, token);
    } else if (repoUrl.includes("gitlab")) {
      return new GitlabRepository(repoUrl, gitCommitHash, token);
    }
    throw new Error("Repository is not supported.");
  }
}

class GithubRepository extends GitRepository {
  private readonly account: string;
  private readonly repoName: string;

  constructor(repoUrl: string, gitCommitHash: string, token?: string) {
    super(repoUrl, gitCommitHash, token);
    const pathParts = this.repo.pathname.slice(1).split("/");
    this.account = pathParts[0];
    this.repoName = pathParts[1];
    if (token) {
      this.httpClient.interceptors.request.use((req) => {
        req.auth = {
          password: this.token,
          username: this.account,
        } as AxiosBasicCredentials;
        return req;
      });
    }
    this.httpClient.interceptors.request.use((req) => {
      req.headers.setAccept("application/vnd.github.object+json");
      req.headers["X-GitHub-Api-Version"] = "2022-11-28";
      return req;
    });
  }

  fileUrl(filepath: string): string {
    return `https://api.github.com/repos/${this.account}/${
      this.repoName
    }/contents/${encodeURIComponent(filepath)}?ref=${encodeURIComponent(
      this.gitCommitHash,
    )}`;
  }

  protected async downloadFileUrl(filepath: string): Promise<string> {
    if (this.token != undefined) {
      return Promise.resolve(
        `https://raw.githubusercontent.com/${this.account}/${this.repoName}/${this.gitCommitHash}/${filepath}`,
      );
    }
    return (await this.httpClient.get(this.fileUrl(filepath))).data[
      "download_url"
    ];
  }
}

class GitlabRepository extends GitRepository {
  private readonly host: string;
  private readonly project: string;

  constructor(repoUrl: string, gitCommitHash: string, token?: string) {
    super(repoUrl, gitCommitHash, token);
    const url = new URL(repoUrl);
    this.host = url.host;
    this.project = url.pathname.slice(1);
    if (token != undefined) {
      this.httpClient.interceptors.request.use((req) => {
        req.headers.setAuthorization(`Bearer ${token}`);
        return req;
      });
    }
  }

  fileUrl(filepath: string): string {
    return `https://${this.host}/api/v4/projects/${encodeURIComponent(
      this.project,
    )}/repository/files/${encodeURIComponent(
      filepath,
    )}?ref=${encodeURIComponent(this.gitCommitHash)}`;
  }

  protected downloadFileUrl(filepath: string): Promise<string> {
    return Promise.resolve(
      `https://${this.host}/api/v4/projects/${encodeURIComponent(
        this.project,
      )}/repository/files/${encodeURIComponent(
        filepath,
      )}/raw?ref=${encodeURIComponent(this.gitCommitHash)}`,
    );
  }
}
